//----------------------------------------------------------------------------
// SVG Viewer - Version 1.0
// A library of SVG renderer
// Copyright (C) 2008 Witold Jaworski
// Contact:	witold.jaworski@airplanes3d.net
//			http://airplanes3d.net
// 
// Contains elements from:
//			AGG - a graphic library (Copyright Maxim Shemanarev, mcseem@antigrain.com)
//          http://antigrain.com
// Uses functions from:
//          Expat - a XML parser library (Copyright James Clark)
// 
// SVG Viewer is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// SVG Viewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SVG Viewer; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
// MA 02110-1301, USA.
//----------------------------------------------------------------------------

#ifndef SVG_BASICS_INCLUDED
#define SVG_BASICS_INCLUDED
#include "agg_array.h"
#include "agg_math_stroke.h"
#include "agg_color_rgba.h"
#include "agg_path_storage.h"
#include "agg_pixfmt_rgba.h"

//acceptable numerical error:
#define SVG_EPSILON 0.000000001 

namespace svg
{

//////////////////////////////////////////////////////////////////////
// The definition of the handle - just an unsigned int
//////////////////////////////////////////////////////////////////////
typedef unsigned handle; //a "handle" to something, stored in the parent node of the
						 //SVG graph (an element, an attribute)
enum { invalid_handle = (unsigned) -1 }; //a constant for the invalid handle

inline bool is_valid(handle hnd) {return (hnd != invalid_handle);}

//////////////////////////////////////////////////////////////////////
// Forward class declarations
//////////////////////////////////////////////////////////////////////
struct reference; //defined in svg_parser.h

//////////////////////////////////////////////////////////////////////
// Shortcut to the types, defined in the AGG library
//////////////////////////////////////////////////////////////////////
typedef agg::trans_affine transform;
typedef agg::rect_base<double> rectangle;
typedef agg::rgba8	rgba8;
typedef agg::rendering_buffer rendering_buffer;
typedef agg::pixfmt_bgra32 pixfmt; 
typedef agg::pixfmt_bgra32_pre pixfmt_pre;  //filters require premultiplied buffers

//////////////////////////////////////////////////////////////////////
// Helper enumerations, used in the attributes
//////////////////////////////////////////////////////////////////////
//Enumeration of the drawing mode (filling shape or stroke)
enum draw_mode_e
{
	none		= 0,	//just to be sure
	solid		= 1,		
	other		= 2		//gradient or pattern
};

//Enumeration of the text writing direction
enum writing_mode_e
{
	lr		= 1,	//chars:left to right, rows:top to bottom
	rl		= 2,	//chars:right to left, rows:top to bottom		
	tb		= 3		//chars:top to bottom, columns: right to left		
};

//Other helpful constants:
/*enum constants
{
	cover_full = agg::cover_full
};*/

//////////////////////////////////////////////////////////////////////
// Helper structures
//////////////////////////////////////////////////////////////////////

struct attribute  
{
	/////////////////////////////////////////////////////////////////
	// attribute: representation of an xml attribute (a single value)
	//			  attribute names are replaced by the enumeration
	/////////////////////////////////////////////////////////////////
    enum kind
    {
			none				= 0, //just to mark an empty value
			style_base			= 0, //just to mark the first index of the style

	//First section: the attributes that can be contained 
	//in the "style" xml attribute:

            fill_flag			= style_base + 1,  //bool. default: false	
            fill_mode			= style_base + 2,  //draw mode: solid color:gradient:pattern
            fill_color			= style_base + 3,  //color. default: white
            fill_opacity		= style_base + 4,  //real. default: 1.0
            fill_rule			= style_base + 5,  //fill rule: nonzero

            stroke_flag			= style_base + 6,  //bool. default: false	
            stroke_mode			= style_base + 7,  //draw mode: solid color:gradient:pattern
            stroke_color		= style_base + 8,  //color. default: black	
            stroke_opacity		= style_base + 9,  //real. default: 1.0	
            
			stroke_width		= style_base + 10,  //length. default: 1.0			
            stroke_linecap		= style_base + 11, //line_cap: butt, square, round
            stroke_linejoin		= style_base + 12, //line_join: miter, round, bevel
            stroke_miterlimit	= style_base + 13, //real: default: 4.0	
			stroke_dasharray	= style_base + 14, //handle to a dash element,
												   //default: invalid_handle (no dash)
			stroke_dashoffset	= style_base + 15, //length. default:0
			font				= style_base + 16, //handle to a font element
			font_size			= style_base + 17, //real. size of the font. Default: 10.0
			writing_mode		= style_base + 18, //writing_mode: lr, rl, tb
			text_anchor			= style_base + 19, //real: 0.0:start, 0.5:middle, 1.0:end 
			kerning				= style_base + 20, //real: negative value = auto. Else - fixed.

			fill_painter		= style_base + 21, //handle to a gradient/pattern element
												   //default: invalid handle (no gradient)
			stroke_painter		= style_base + 22, //handle to a gradient/pattern element
												   //default: invalid handle (no gradient)
	//helper marks:
			style_max			= style_base + 23, //just to mark the last index of the style
			style_mask			= 0x00FF,

	//Other kind of attributes
			dash_span			= style_mask + 1, //real (used by svg::dash element)
			filter_ref			= style_mask + 2, //handle (used in shape::override())				
			mask_ref			= style_mask + 3, //handle (used in shape::override())				
			clip_path_ref		= style_mask + 4, //handle (used in shape::override())				
			opacity_ref			= style_mask + 5, //handle (used in shape::override())				
			display				= style_mask + 6, //bool (used in element::override())				
    //      name2				= style_max + 7,
	//	...
    } type;


	union value
	{
		int					integer;
		float				real;
		bool				boolean;
		agg::int8u			rgba[4]; //a substitute to store a color data
		agg::line_join_e	line_join;
		agg::line_cap_e		line_cap;
		draw_mode_e			draw_mode;
		writing_mode_e		writing_mode;
		agg::filling_rule_e	filling_rule;
		handle				id;
	} val;

	//Just to enumerate the color memebers in the attribute::value rgba array
	enum colorplane
	{
		B = 0,		//Blue
		G = 1,		//Green
		R = 2,		//Red
		A = 3		//Alpha
	};

	//////////////////////////////////////////////////////////////////////
	// helper functions
	// no checking of the proper match type to value,so far!			  
	//////////////////////////////////////////////////////////////////////
	//bool has_value() {return type != attribute::none;}

	agg::rgba8 val2color() 
	{	return agg::rgba8 (val.rgba[R],val.rgba[G],val.rgba[B],val.rgba[A]); } 

	double val2double() {return double(val.real);} 

	//Helper function, that returns a fill_mode or stroke_mode constants
	//Arguments
	//	type:	fill_flag or stroke_flag
	//
	static kind mode(kind type) 
					{return (type == fill_flag)? fill_mode : stroke_mode;}
	//Helper function, that returns a fill_color or stroke_color constants
	//Arguments
	//	type:	fill_flag or stroke_flag
	//
	static kind color(kind type) 
					{return (type == fill_flag)? fill_color : stroke_color;}
	//Helper function, that returns a fill_painter or stroke_painter constants
	//Arguments
	//	type:	fill_flag or stroke_flag
	//
	static kind painter(kind type) 
					{return (type == fill_flag)? fill_painter : stroke_painter;}
	//Helper function, that returns a fill_opacity or stroke_opacity constants
	//Arguments
	//	type:	fill_flag or stroke_flag
	//
	static kind alpha(kind type) 
					{return (type == fill_flag)? fill_opacity : stroke_opacity;}

	//////////////////////////////////////////////////////////////////////
	// attribute constructors: 
	// no checking of the proper match type to the value,so far!			  
	//////////////////////////////////////////////////////////////////////
	attribute():type(none) {val.integer = 0;}

	attribute(kind atype, int avalue) : type(atype) {val.integer = avalue;}
	attribute(kind atype, handle avalue) : type(atype) {val.id = avalue;}
	attribute(kind atype, float avalue) : type(atype) {val.real = avalue;}
	attribute(kind atype, double avalue) : type(atype) {val.real = (float) avalue;}
	attribute(kind atype, bool avalue) : type(atype) {val.boolean = avalue;}
	attribute(kind atype, agg::line_join_e avalue) : type(atype) {val.line_join = avalue;}
	attribute(kind atype, agg::line_cap_e avalue) : type(atype) {val.line_cap = avalue;}
	attribute(kind atype, draw_mode_e avalue) : type(atype) {val.draw_mode = avalue;}
	attribute(kind atype, writing_mode_e avalue) : type(atype) {val.writing_mode = avalue;}
	attribute(kind atype, agg::filling_rule_e avalue) : type(atype) {val.filling_rule = avalue;}
	attribute(kind atype, agg::rgba8 avalue) : type(atype) 
		{val.rgba[R] = avalue.r; val.rgba[G] = avalue.g;
		 val.rgba[B] = avalue.b; val.rgba[A] = avalue.a;}
		
}; //attribute
/*
When you want to implement another style attribute:
	1.	Extend the definition of the svg::attribute structure, above,
		adding to it a new kind, and - optionally - a new data type
		to the union value. When you add a new data type - usually
		a new enum - do not forget to create a new constructor for it.

	2.	Update the parser::store_as_style() function in svg_parser.cpp,
		to recognize and properly create the attribute;

	3.	Use it somewhere in the code:
		-	place a default value in the styleset constructor
			(svg_styleset.h);
		-	use it to update an object, usually in one of the 
			styleset::apply*() functions (svg_styleset.h)
*/
//////////////////////////////////////////////////////////////////////
// Definitions of the main storages: geometry, attributes, elements
//////////////////////////////////////////////////////////////////////
typedef agg::path_base<agg::vertex_block_storage<float> > path_storage;
typedef agg::pod_bvector<attribute> attr_storage;
typedef agg::pod_bvector<reference> element_storage;

//////////////////////////////////////////////////////////////////////
// Some other useful stuff:
//////////////////////////////////////////////////////////////////////
//Helper function - just to compare two rectangles
bool is_equal(const rectangle &r1, const rectangle &r2);

//Helper function - checks, if the rectangle has zero width or height 
bool has_no_area(const rectangle &rec); 

//Helper function - checks if the transformation is just a 
//plain translation, without scaling, shear or rotation
//Arguments
//	trans:	transformation to check
bool is_translation_only(transform trans);

//Helper function - checks if the transformation is empty
//Arguments
//	trans:	transformation to check
bool is_empty(transform trans);

//Returns true, when the inner rectangle is inside outer
//Arguments:
//	inner:	the presumably smaller rectangle
//	outer:	the presumably larger rectangle
bool is_inside(const rectangle &inner, const rectangle &outer);

bool is_equal(rgba8 &c1, rgba8 &c2);

} //namespace svg
#endif //SVG_BASICS_INCLUDED

