//----------------------------------------------------------------------------
// SVG Viewer - Version 1.0
// A library of SVG renderer
// Copyright (C) 2008 Witold Jaworski
// Contact:	witold.jaworski@airplanes3d.net
//			http://airplanes3d.net
// 
// Contains elements from:
//			AGG - a graphic library (Copyright Maxim Shemanarev, mcseem@antigrain.com)
//          http://antigrain.com
// Uses functions from:
//          Expat - a XML parser library (Copyright James Clark)
// 
// SVG Viewer is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// SVG Viewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SVG Viewer; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
// MA 02110-1301, USA.
//----------------------------------------------------------------------------

#ifndef SVG_ALPHA_MASK_INCLUDED
#include "agg_alpha_mask_u8.h"

#include "svg_filter_primitive.h"
#define SVG_ALPHA_MASK_INCLUDED

namespace svg
{
//////////////////////////////////////////////////////////////////////
// fe_alpha_mask: a base class for the fe_mask and fe_clip.
// It is used for alpha composition of two image buffers: input and the 
// mask.
//////////////////////////////////////////////////////////////////////
class fe_alpha_mask : public filter_primitive
{
	public:
		//Just a shortcut for the color components (for the children classes)
		enum component
		{
			R = pixfmt::order_type::R,
			G = pixfmt::order_type::G,
			B = pixfmt::order_type::B,
			A = pixfmt::order_type::A
		};

	public:
		//////////////////////////////////////////////////////////////////////
		// constructor and overriden virtual methods
		//////////////////////////////////////////////////////////////////////

		fe_alpha_mask(element *parent, unsigned level): filter_primitive(parent, level)
		{}

														 
	protected:

		//Template of the static constructor - to create the child classes in the
		//commit() method of svg::clip_path and svg::mask classes.
		//Arguments:
		//	prs:		parser;
		//	elm:		a new mask, just created with new()
		//	commited:	true, when the created element has to be already commited
		static element *create(parser &prs, fe_alpha_mask *elm, bool commited = true)
		{
			if (elm) 
			{
				prs.store(*elm,"",false,false); //store unnamed element
				if (commited) elm->commit(prs);
			}
			return elm;
		}

		typedef agg::comp_op_rgba_dst_in<rgba8, pixfmt::order_type> blend_func;

		template <class AlphaMask>
		void mask(req_filter &req, AlphaMask &amask)
		{
			pixfmt_pre &dst = req.get("");

			agg::int8u *p, cover; //destination alpha, mask alpha
			int x,y;
			int width = req.width();
			int height = req.height();

			for (y = 0; y < height; y++)
			{
				for (x =0; x < width; x++)
				{
					cover = amask.pixel(x,y);
					p	  = dst.pix_ptr(x,y);
					
					switch (cover)
					{
					case 0:
						p[A] = 0; //skip this pixel
						break;
					case agg::cover_full:
						//no change in dst - full copy 
						break;
					default: //when p[A] < cover_full and cover < cover_full
							 //in fact, this case is not so often:
							 //combine this pixel's alpha value.
						if (p[A])
						{
							blend_func::blend_pix(p,0,0,0,cover,agg::cover_full);
						}
						
					}//switch
				}//for x
			}//for y
		}

}; //fe_alpha_mask

//////////////////////////////////////////////////////////////////////
// fe_mask: performs an alpha composition of its mask with the result
//////////////////////////////////////////////////////////////////////
class fe_mask : public fe_alpha_mask
{
	public:
		//////////////////////////////////////////////////////////////////////
		// constructor and overriden virtual methods
		//////////////////////////////////////////////////////////////////////

		fe_mask(element *parent, unsigned level): fe_alpha_mask(parent, level)
		{}

		//Static constructor - used by the svg::clip_path, svg::mask to create and store 
		//a new svg::tspan element to handle its direct content (in the commit() method)
		//Arguments:
		//	prs:		parser;
		//	parent:		the text element, that will be the parent of this element
		//	level:		the level of nesting in the source XML tree (just 
		//				parent->level()+1 here)
		//	commited:	true, when the created element has to be already commited
		static element *create(parser &prs, element *parent, unsigned level, bool commited = true)
		{
			return fe_alpha_mask::create(prs,new fe_mask(parent,level),commited);
		}

		virtual const char * tag() {return "fe_mask";}; 
														 
	public:
		struct color_transform
		{
			static unsigned calculate(const agg::int8u* p) 
			{ 
				if(p[A] == 0)
				{
					return 0;
				}
				else
				{
					return (((p[R]*77 + p[G]*150 + p[B]*29) >> 8) * p[A])>>8; 
				}
			}
		};//color_transform

		typedef agg::alpha_mask_u8<pixfmt::pix_width,0,color_transform> alpha_mask;

		//override of the base class method:
		virtual handle perform(req_filter &req)
		{
			filter_primitive::perform(req); //always call the base, first!
			alpha_mask amask;
			amask.attach(req.get_buf("#mask"));
			mask(req,amask);
			return skip(req);
		}
}; //fe_mask
//////////////////////////////////////////////////////////////////////
// fe_clip: performs a simple elimination of pixels 
//////////////////////////////////////////////////////////////////////
class fe_clip : public fe_alpha_mask
{
	public:
		//////////////////////////////////////////////////////////////////////
		// constructor and overriden virtual methods
		//////////////////////////////////////////////////////////////////////

		fe_clip(element *parent, unsigned level): fe_alpha_mask(parent, level)
		{}

		//Static constructor - used by the svg::clip_path, svg::mask to create and store 
		//a new svg::tspan element to handle its direct content (in the commit() method)
		//Arguments:
		//	prs:		parser;
		//	parent:		the text element, that will be the parent of this element
		//	level:		the level of nesting in the source XML tree (just 
		//				parent->level()+1 here)
		//	commited:	true, when the created element has to be already commited
		static element *create(parser &prs, element *parent, unsigned level, bool commited = true)
		{
			return fe_alpha_mask::create(prs,new fe_clip(parent,level),commited);
		}

		virtual const char * tag() {return "fe_clip";}; 
														 
	public:
		struct color_transform
		{
			static unsigned calculate(const agg::int8u* p) 
			{ 
				return p[A];
			}
		};//color_transform

		typedef agg::alpha_mask_u8<pixfmt::pix_width,0,color_transform> alpha_mask;

		//override of the base class method:
		virtual handle perform(req_filter &req)
		{
			filter_primitive::perform(req); //always call the base, first!
			alpha_mask amask;
			amask.attach(req.get_buf("#clip"));
			mask(req,amask);
			return skip(req);
		}
}; //fe_clip

}; //namespace svg
#endif //SVG_ALPHA_MASK_INCLUDED

