//----------------------------------------------------------------------------
// SVG Viewer - Version 1.0
// A library of SVG renderer
// Copyright (C) 2008 Witold Jaworski
// Contact:	witold.jaworski@airplanes3d.net
//			http://airplanes3d.net
// 
// Contains elements from:
//			AGG - a graphic library (Copyright Maxim Shemanarev, mcseem@antigrain.com)
//          http://antigrain.com
// Uses functions from:
//          Expat - a XML parser library (Copyright James Clark)
// 
// SVG Viewer is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// SVG Viewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SVG Viewer; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
// MA 02110-1301, USA.
//----------------------------------------------------------------------------

#ifndef SVG_PATTERN_INCLUDED
#include "agg_span_allocator.h"
#include "agg_span_interpolator_linear.h"
#include "agg_span_image_filter_rgba.h"
#include "agg_span_gradient.h"

#include "svg_element.h"
#include "svg_painter.h"

#define SVG_PATTERN_INCLUDED

namespace svg
{
//////////////////////////////////////////////////////////////////////
// repeat_interpolator: a helper class: interpolator, which oscilates 
// the coordinates within area that has left upper corner at (-off_x, -off_y)
// and giwen width and height
//////////////////////////////////////////////////////////////////////
	class repeat_interpolator : public agg::span_interpolator_linear<>
	{
	private:
		typedef agg::span_interpolator_linear<> base_type;
	public:
		//Constructor:
		//Arguments
		//	trans:	defines a transformation form the final render coordinates
		//			to the coordinates of the source picture. In this final coo-
		//			rdinate system source picture has always its upper left 
		//			corner at 0,0.
		repeat_interpolator( transform &trans): base_type(trans) {}

		//Initialization:
		//Arguments:
		//	off_x:	distance from the left edge of the picture to the left edge
		//			of repeated range (can be nonzero)
		//	off_x:	distance from the top edge of the picture to the top edge
		//			of repeated range (can be nonzero)
		//	width:	horizontal size of the repeated range (can be different 
		//			the the size of the source picture)
		//	height:	vertical size of the repeated range (can be different 
		//			the the size of the source picture)
		void set(double off_x, double off_y,	double width, double height)
		//All values in the interpolator are processed in the subpixel accuracy.
		//We have store the received values with just the same setting
		{
			m_off_x = agg::iround(off_x * base_type::subpixel_scale);
			m_off_y = agg::iround(off_y * base_type::subpixel_scale);
			m_width = agg::iround(width * base_type::subpixel_scale);
			m_height = agg::iround(height * base_type::subpixel_scale);
		}

		
		//Returns the actual coordinates of the interpolator:
		//Arguments:
		//	x:	pointer to a x coordinate value;
		//	y:	pointer to a y coordinate value;
		void coordinates(int *x, int *y) const
		{
			base_type::coordinates(x, y);
			*x = (*x + m_off_x) % m_width;
			if (*x < 0) *x += m_width;
			*y = (*y + m_off_y) % m_height;
			if (*y < 0) *y += m_height;
		}

	protected:
		int m_off_x, m_off_y; //placement of the upper left picture
							  //corner within repeated range;
		int m_width, m_height;//size of the repeated range 
	};//repeat_interpolator
//////////////////////////////////////////////////////////////////////
// pattern_pixfmt: a child of pixfmt, specialized for handling patterns
//               (it can be used many times, each with different opacity)			 
//////////////////////////////////////////////////////////////////////
class pattern_pixfmt: public pixfmt
{
	public:

		enum cnst 
		{
			def_size = 256 //default size of the picture (def_szie x def_size)
						   //when pattern content units are relative to object
						   //bounding box, which uses the pattern. 
						   //(patternContentUnits = "objectBoundingBox")
		};

	public:

		//Constructor: like in the pixfmt. The rendering buffer will be
		//assigned later on - in the initialize() method:
		pattern_pixfmt(): pixfmt(), m_alpha(NULL),m_opacity(1.0) {}

		//A copy constructor, implemented nowhere (just to block such operation
		//at least on the linker level - it should never happen in the code)
		pattern_pixfmt(pattern_pixfmt &ppf); 

		//Destructor: we have to clear the eventual alpha buffer:
		~pattern_pixfmt()
		{
			if (m_alpha) delete [] m_alpha;
			m_alpha = NULL;
		}

		//Initializes the object:
		//
		void initialize(rendering_buffer &rb)
		{
			attach(rb);
			if (m_alpha) delete [] m_alpha;
			m_alpha = new agg::int8u[rb.width()*rb.height()];
			if (m_alpha)
			{
				unsigned row_width = width();
				unsigned x,y;
				for (y = 0; y < height(); y++)
				{
					for(x = 0; x < row_width; x++)
					{
						m_alpha[y*row_width + x] = pixel(x,y).a; //copy original opacity
					}//end for
				}//end for
			}

		}

		bool is_valid() {return (m_alpha != NULL);}

		//Prepares the picture for the use:
		//Arguments:
		//	opacity:	(0.0 ... 1.0) - the overall opacity of the object
		void set(double opacity)
		{
			if (m_opacity != opacity && m_alpha)
			{
				unsigned row_width = width();
				unsigned x,y;
				for (y = 0; y < height(); y++)
				{
					for(x = 0; x < row_width; x++)
					{
						rgba8 p = pixel(x,y);	//take a pixel from the source:
						p.a = m_alpha[y*row_width + x];		//restore original opacity
						p.opacity(p.opacity()*opacity); //multiple by the global opacity
						copy_pixel(x, y, p);	//put it back to the source
					}//end for
				}//end for
				m_opacity = opacity;
			}
		}//end set


	protected:
		//Why we have to preserve the original alpha values in an buffer,
		//instead of dividing by previous opacity before multiply by the
		//new one? Because sometimes the previous opacity can be 0.0.
		double		m_opacity;	//last used value of the "global opacity"
		agg::int8u	*m_alpha;	//buffer for the original alpha values
								//Can be NULL, when not assigned to any
								//"real" picture. 
};//pattern_pixfmt
//////////////////////////////////////////////////////////////////////
// pattern: an element of the SVG graph, 
//			defines a pattern for fillin or stroking
//////////////////////////////////////////////////////////////////////
class pattern : public painter 
{

	public:
		//////////////////////////////////////////////////////////////////////
		// constructor and overriden virtual methods
		//////////////////////////////////////////////////////////////////////

		pattern(element *parent, unsigned level): painter(parent, level),
		m_x(0.0), m_y(0.0), m_width(0.0), m_height(0.0),
		m_box(0,0,-1,-1), //this box is not valid (intentionally => not initialized)
		m_bounds(0,0,-1,-1), //this box is not valid (intentionally => not initialized)
		m_picture(NULL),  //can receive value in the initialize() method
		m_interpolator(m_actual_trans),
		m_generator(m_pixels, rgba8(0,0,0,0), m_interpolator)
		{}

		~pattern() //remove the assigned picture buffer:
		{
			if(m_picture) delete [] m_picture;
			m_picture = NULL;
		}

		virtual const char * tag() {return "pattern";};

		virtual void on_attribute(parser &prs, const char * name, const char *value)
		{
			if (strcmp(name, "patternUnits") == 0) 
			{	
				if (strcmp(value, "objectBoundingBox")==0) m_relative_units = 1;
				if (strcmp(value, "userSpaceOnUse")==0) m_relative_units = 0;
			}
			else
			if (strcmp(name, "patternContentUnits") == 0) 
			{	
				if (strcmp(value, "objectBoundingBox")==0) m_relative_content_units = 1;
				if (strcmp(value, "userSpaceOnUse")==0) m_relative_content_units = 0;
			}
			else
			if (strcmp(name, "patternTransform") == 0) 
			{
				m_extra_trans = prs.parse_transform(value);
				m_extra_trans.invert();
			}
			else
			if (strcmp(name, "x") == 0) 
			{	
				m_x = parse_coordinate(prs,value);
			}
			else
			if (strcmp(name, "y") == 0) 
			{	
				m_y = parse_coordinate(prs,value);
			}
			else
			if (strcmp(name, "width") == 0) 
			{	
				m_width = parse_coordinate(prs,value);
			}
			else
			if (strcmp(name, "height") == 0) 
			{	
				m_height = parse_coordinate(prs,value);
			}
			else
			{//if the style is not recognized:
				painter::on_attribute(prs,name,value);
			}
		}
		
		virtual bool is_of(const char *cls) 
		{
			if(strcmp("pattern",cls) == 0)
			{
				return true;
			}
			else //it is not your name? Make sure, asking your parent :)
				return painter::is_of(cls);
		}

		//Method required by AGG to be used in the render_scanlines_aa() method:
		//to be overriden in the children classes:
		virtual void generate(rgba8* span, int x, int y, unsigned len)
		{
			m_generator.generate(span, x, y, len);
		}

		//This method is called inside req_render::perform method,
		//just before the render_scanlines_aa
		//Arguments:
		//	gtrans:	"global" transformation of the shape (from its local 
		//			coordinates to the picture coordinates
		//	opacity:overall opacity of the gradient/pattern. (0.0 .. 1.0)
		//			used in the child classes
		virtual void set(transform &gtrans, double opacity)
		{

			{
				double scale = m_relative_content_units? pattern_pixfmt::def_size : 1.0;
				
				//width, height: have to be expressed in the pixel coordinates:
				rectangle box = m_box;
				transform to_pixels = m_unit_trans;
				to_pixels *= m_internal_trans;
				to_pixels.transform(&(box.x1),&(box.y1));
				to_pixels.transform(&(box.x2),&(box.y2));
				
				double xoffs = m_bounds.x1*scale;
				double yoffs = m_bounds.y1*scale;
				double width  = fabs(box.x2 - box.x1); //to_pixels transfrom already scaled it
				double heigth = fabs(box.y2 - box.y1); //to_pixels transfrom already scaled it

				m_interpolator.set(xoffs, yoffs, width, heigth);
			}

			m_pixels.set(opacity);
			painter::set(gtrans, opacity);
		}

				//Sets the unit transformation (when unit mode is objectBoundingBox)
		//Arguments:
		//	box:	the reference bounding box (use shape::bounding_box() here)
		//Should be called for gradients which have relative units, only.
		//Call it BEFORE req_render::perform method
		virtual void set(rectangle &box)
		{
			m_unit_trans = transform();
			m_box = this->box(); //reset the m_box
			
			transform box_trans = transform(); //relative transform
			box_trans.scale(box.x2 - box.x1, box.y2 - box.y1); 
			box_trans.translate(box.x1, box.y1);

			if (m_relative_units == 1) //m_box is always in the absolute units:
			{
				box_trans.transform(&(m_box.x1),&(m_box.y1));
				box_trans.transform(&(m_box.x2),&(m_box.y2));
			}

			if (m_relative_content_units == 1) 
			{
				m_unit_trans = box_trans;
			}
			else
			{
				m_unit_trans.translate(m_box.x1, m_box.y1);
			}

			m_unit_trans.invert();
		}

		virtual void reset() //called, when there are no relatives:
		{
			m_unit_trans = transform();
			m_box = box(); //reset the m_box

			m_unit_trans.translate(m_box.x1, m_box.y1);
			m_unit_trans.invert();
		}
	public:
		//////////////////////////////////////////////////////////////////////
		// Simple methods for copying the content of the referenced gradient
		// (see apply() method)
		//////////////////////////////////////////////////////////////////////

		//returns the pattern box (because m_box can be changed at each set())
		//we are transferring the 
		rectangle box() 
		{
			if (m_width != 0.0 && m_height != 0.0) //initailized?
				return rectangle(m_x, m_y, m_x + m_width, m_y+m_height);
			else
				return rectangle(m_x, m_y, m_x-1, m_y -1); //not valid!
		}

		//Returns the render buffer: check its stride(). When it is nonzero,
		//it should be used - assign it to the m_pixels in the use() 
		//(Such rendering buffer is attached to a valid m_picture.)
		rendering_buffer &buffer(){return m_rb;} 

		//Returns the "natural box" - size of the pattern elements 
		//expressed in the nominal coordiantes
		rectangle bounding_box()
		{
			if (!m_bounds.is_valid())
			{
				req_measure measure;
				req_prepare prepare(request::prepare_measure, measure); 
				//make the measure request usable:
				to_root(prepare); 
				if(measure.can_query())
				{
					measure.trans.reset();//clear the internal transform
										  //because bounding box is expressed										  //INSIDE the element's frame
					forward(measure);
				}//end if
				m_bounds = measure.bounds;
			}
			return m_bounds;
		}//natural_box

		//////////////////////////////////////////////////////////////////////
		// Other methods and fields
		//////////////////////////////////////////////////////////////////////

	protected:
		//This method is called within painter::initialize, if the m_href is valid.
		virtual void use(request &req, painter *src)
		{
			painter::use(req, src);

			if (!src->is_of("pattern")) return; //something wrong - this is not a gradient

			pattern *href = (pattern *) src;

			if (href->box().is_valid()) //copy box, if valid 
			{
				m_box = href->box(); //because m_box is overriden in each set() call
									 //we have to preserve received results in the 
				m_x = m_box.x1;		 //m_x, m_y, m_width, m_height parameters
				m_y = m_box.y1;
				m_width = m_box.x2 - m_box.x1;
				m_height = m_box.y2 - m_box.y1;
			}

			if (!m_bounds.is_valid() && href->bounding_box().is_valid()) //copy box, if valid 
			{
				m_bounds = href->bounding_box(); 
			}

			if (!buffer().stride() && href->buffer().stride()) //copy picture, if valid
			{
				m_pixels.initialize(href->buffer());
			}
			
		}	
	

		//Initializes the pattern, when the object receives a reqest::initialize
		//Arguments:
		//	req:	req::initialize - passed to give the access to the element and
		//			attribute storage.	
		virtual void initialize(request &req)
		{
			painter::initialize(req);

			if (! m_initialized) return;

			//form the internal transformation:
			rectangle rect = bounding_box();
			m_internal_trans = pixels_to_nominal(rect); 
			m_internal_trans.invert(); //final inversion: expected by picture::set()
			
			if(!m_pixels.is_valid() && elements() > 0) //this 
			{
				//attach and render content of the m_rb
				//(it also sets up the m_internal_transform, and m_picture:)
				if(prepare_picture(m_rb, rect))
				{
					render_pixels(m_rb,m_internal_trans); //it assigns m_rb to m_pixels
					m_pixels.initialize(m_rb); //initializes the alpha channel
				}
			}

			m_initialized = (m_pixels.stride() > 0); //this object is valid, 
											   //when pixels are in place (stride is non-zero)
		}

		//Returns a proper scaling from nominal coordinates to pixel coordinates
		//Arguments:
		//	nbox:		natural extents of the pattern picture, nominal coordinates
		double pixel_scale(rectangle nbox)
		{
			if (m_relative_content_units)
			{
				double width = (nbox.x2 - nbox.x1); //width of the nbox: used for relative sizing
				if (width > 1.0) width = 1.0; 
				if (width > SVG_EPSILON)
				{
					return pattern_pixfmt::def_size/width;
				}
				else return SVG_EPSILON; //nonzero, to not degenerate transformation
			}
			else return 1.0;
		}
		//Prepares the internal transformation (from pixel to nominal coordinates)
		//Arguments:
		//	nbox:		natural extents of the pattern picture, nominal coordinates
		transform pixels_to_nominal(rectangle nbox)
		{
			transform result = transform();
			if (nbox.is_valid())
			{
				result.scale(1.0 / pixel_scale(nbox));
				result.translate(nbox.x1, nbox.y1);
			}
			return result;
		}
		
		//Prepares the m_picture
		//Arguments:
		//	rbuf:		rendering buffer, which will contain the picture bytes
		//	nbox:		natural extents of the pattern picture, nominal coordinates
		bool prepare_picture(rendering_buffer &rbuf, rectangle &nbox)
		{
			if (nbox.is_valid())
			{
				double scale = pixel_scale(nbox);
				unsigned width = agg::uround((nbox.x2 - nbox.x1)*scale);
				unsigned height = agg::uround((nbox.y2 - nbox.y1)*scale);
				m_picture = new agg::int8u [width * height * pattern_pixfmt::pix_width];
				if (m_picture) 
				{
					rbuf.attach(m_picture, width, height, width * pattern_pixfmt::pix_width);
					m_pixels.attach(rbuf);
				}
				return (m_picture != NULL);
			}
			else
			{
				return false;
			}
		}

		//Renders the pattern into given rendering buffer
		//Arguments (passed this way for more readable code, only!):
		//	rbuf:	rendering buffer - the target of the rendering
		//	trans:	transformation from nominal coordiantes to pixel coordinates
		void render_pixels(rendering_buffer &rbuf, transform trans)
		{
			rectangle clip(0.0, 0.0, m_pixels.width(), m_pixels.height());
			req_render render(m_pixels, clip, rgba8(0,0,0,0));
			req_prepare prepare(request::prepare_render, render);
			//make the measure request usable:

			to_root(prepare); 
			if(render.can_query())
			{
				render.trans = trans; 
				forward(render);
			}//end if
		}//render pixels
		
	public:
		typedef agg::span_image_filter_rgba_bilinear_clip<pattern_pixfmt, 
                                          repeat_interpolator> span_generator;
	protected:
		//////////////////////////////////////////////////////////////////////
		// Instance additional fields
		//////////////////////////////////////////////////////////////////////
		//helper attributes, used in set() method to create a m_box.
		//m_box can be enlarged when patternUnits = objectBoundingBox
		double m_x, m_y;			
		double m_width, m_height;
		
		//Other fields
		rectangle m_box;	//the elementary box, that has to be repeated.
							//(Created from x, y, width, height properties)
							//Its size may not match the size of the picture.
							//(x1, y1) may not be at the source picture (0,0)
							//Always in the non-relative coordinates!

		rectangle m_bounds; //the bounds of the pattern, in the nominal 
							//coordinates (depending on the m_relative_units
							//it can also be the relative size)

		agg::int8u *m_picture;	//Source buffer. May be NULL, for the patterns that
								//do not have the content elements (and play the role
								//of a reference to another pattern)
		rendering_buffer m_rb;  //rendering buffer, used by the m_pixels. 

		pattern_pixfmt m_pixels;//the required form of the pattern picture - Can be 
								//attached to a rendering buffer from another <pattern> 
								//element, in a pattern that plays the reference
								//role, only.
		
		//The image generator requires an interpolator:
		repeat_interpolator m_interpolator;

		//And finally: the pattern object itself:
		span_generator m_generator;
}; //pattern

}; //namespace svg
#endif //SVG_PATTERN_INCLUDED

