//----------------------------------------------------------------------------
// SVG Viewer - Version 1.0
// A library of SVG renderer
// Copyright (C) 2008 Witold Jaworski
// Contact:	witold.jaworski@airplanes3d.net
//			http://airplanes3d.net
// 
// Contains elements from:
//			AGG - a graphic library (Copyright Maxim Shemanarev, mcseem@antigrain.com)
//          http://antigrain.com
// Uses functions from:
//          Expat - a XML parser library (Copyright James Clark)
// 
// SVG Viewer is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// SVG Viewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SVG Viewer; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
// MA 02110-1301, USA.
//----------------------------------------------------------------------------

#ifndef SVG_STYLESET_INCLUDED
#define SVG_STYLESET_INCLUDED
#include "svg_basics.h"

namespace svg
{
//////////////////////////////////////////////////////////////////////
// svg::styleset: a full set of all defined style attributes.
// It content matches the actual "style state" - all these
// style settings inherited from the containers, or reset
// at the single element level.
// Instances of this class are attached to the svg::req_render request. 
// Together with the other data, they are forwarded to all elements 
// are rendered into the final image.
//////////////////////////////////////////////////////////////////////
struct styleset  
{
    public:
		//Constructor - fills the styleset with default values
		//Ultimately, it can leave some items with the empty type
		//(attribute::none) - to mark, that they are NOT set (yet?)
		styleset()
		{
			//Default values (add here ALL implemented ones, with moreless "inactive" value)
			set(attribute(attribute::fill_flag,true)); //fill is on, by default
			set(attribute(attribute::fill_mode,solid));
			set(attribute(attribute::fill_color,rgba8(0,0,0)));
			set(attribute(attribute::fill_opacity,(float) 1.0));
			set(attribute(attribute::fill_rule,agg::fill_non_zero));

			set(attribute(attribute::stroke_flag,false));
			set(attribute(attribute::stroke_mode,solid));
			set(attribute(attribute::stroke_color,rgba8(0,0,0)));
			set(attribute(attribute::stroke_opacity,(float) 1.0));

			//Some files do not specify these attributes - we have to imply them!
			set(attribute(attribute::stroke_width,(float) 1.0)); 
			set(attribute(attribute::stroke_linecap,agg::round_cap));
			set(attribute(attribute::stroke_linejoin,agg::round_join));
			set(attribute(attribute::stroke_miterlimit,(float) 4.0));
			set(attribute(attribute::stroke_dasharray,invalid_handle));
			set(attribute(attribute::stroke_dashoffset,(float) 0.0));
			set(attribute(attribute::font,invalid_handle));
			set(attribute(attribute::font_size,(float) 72.0));
			set(attribute(attribute::text_anchor,(float) 0.0));
			set(attribute(attribute::writing_mode,lr));

			//kerning: an adjustment of the space between characters
			//that brings the shapes into better distance (for example -
			//"a" under "T" in a phrase "Ta".
			//negative value means kerning from the font definition
			//(called here: "dynamic"). 0 or higher value - fixed distance
			//between characters (called here a "static" kerning)
			set(attribute(attribute::kerning,-1.0));
			//According SVG specification default kerning is a dynamic kerning,
			
			set(attribute(attribute::fill_painter,invalid_handle));
			set(attribute(attribute::stroke_painter,invalid_handle));
		}//styleset::styleset

		//Destructor - nothing special to do
        virtual ~styleset() {};

		//Resets style to render just a plain contours, without any strokes, 
		//nor filters. Used by clip filters to prepare their masks
		//Placed here to not omit it when a new style element will be implemented
		void to_plain_contour()
		{
			set(attribute(attribute::fill_flag,true));
			set(attribute(attribute::fill_mode,solid));
			set(attribute(attribute::fill_color,rgba8(1,1,1)));
			set(attribute(attribute::stroke_flag,false));
		}

		//Returns true, when this styleset contains the attr attribute type
		//Arguments:
		//	attr:	requested attribute type
		bool has(attribute::kind attr) 
		{
			if ((attr >= attribute::style_base) && (attr <= attribute::style_max))
				return m_content[attr].type != attribute::none;
			else return false;
		}

		//Returns the value of given attribute
		//Arguments:
		//	attr:	requested attribute type
		attribute of(attribute::kind attr) 
		{
			if (has(attr)) return m_content[attr];
			else return attribute(); //Error situation - return an empty value
		}

		//Updates the value of given attribute
		//Arguments:
		//	attr:	a new attribute
		//Returns true, when update suceeded.
		bool set(attribute attr) 
		{
			if ((attr.type >= attribute::style_base) && (attr.type <= attribute::style_max))
			{
				m_content[attr.type] = attr; 
				return true;
			}
			else return false; //Error situation 
		}
		

		//Updates the styleset, using given range of attributes, stored
		//within the request object
		//Arguments:
		//	src:		an attribute container - usually the request object
		//	from_id:	handle to the first of attributes, that should be set.
		//	to_id:		handle to the last of attributes
		//Returns true, if dataset has been changed (even by a single attribute)
        template<class AttributeContainer> 
		bool update(AttributeContainer &src, handle from_id, handle to_id)
		{
			bool result = false;
			handle id;
			if (to_id == invalid_handle) return result; //no attributes

			for(id = from_id; id <= to_id; id++)
			{
				attribute *attr = src.get_attribute(id);
				if(attr) result |= set(*attr);
			}
			return result;
		}//styleset::update


		//Applies stroke attributes to a vertex source (a conv_stroke<> template)
		//Arguments:
		//	vc:			the vertex converter (one of conv_stroke<>) to be modified
        template<class VertexConverter> 
		void apply_stroke(VertexConverter &vc)
		{
			vc.width(of(attribute::stroke_width).val2double());
			vc.line_join(of(attribute::stroke_linejoin).val.line_join);
			vc.line_cap(of(attribute::stroke_linecap).val.line_cap);
			vc.miter_limit(of(attribute::stroke_miterlimit).val2double());
			vc.inner_join(agg::inner_round);
		}

		//Returns resulting opacity for the fills
		//Arguments:
		inline double fill_opacity()
		{
			return of(attribute::fill_opacity).val2double();
		}
		
		//Returns resulting opacity for the strokes
		//Arguments:
		inline double stroke_opacity()
		{
			return of(attribute::stroke_opacity).val2double();
		}

		//Applies proper fill color to a solid renderer
		//Arguments:
		//	rnd:		the renderer
		template<class RendererSolid>
		void apply_fill_color(RendererSolid &rnd)
		{
			rgba8 color = of(attribute::fill_color).val2color();
			color.opacity(fill_opacity());
			rnd.color(color);
		}

		//Applies proper stroke color to a solid renderer
		//Arguments:
		//	rnd:		the renderer
		template<class RendererSolid>
		void apply_stroke_color(RendererSolid &rnd)
		{
			rgba8 color = of(attribute::stroke_color).val2color();
			color.opacity(stroke_opacity());
			rnd.color(color);
		}

    protected:
		//the table, that contains single instance of each defined attribute.
		//It should at least the default value for each one.
		attribute m_content[attribute::style_max+1];
}; //styleset


}; //namespace svg
#endif //SVG_STYLESET_INCLUDED

